/* serverside.c   Handles the server side of dopewars                   */
/* Copyright (C)  1998-9  Ben Webb                                      */
/*                Email: ben@bellatrix.pcl.ox.ac.uk                     */
/*                WWW: http://bellatrix.pcl.ox.ac.uk/~ben/dopewars/     */

/* This program is free software; you can redistribute it and/or        */
/* modify it under the terms of the GNU General Public License          */
/* as published by the Free Software Foundation; either version 2       */
/* of the License, or (at your option) any later version.               */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program; if not, write to the Free Software          */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               */
/*                   MA  02111-1307, USA.                               */


#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include "serverside.h"
#include "dopewars.h"
#include "message.h"
#include "AIPlayer.h"

#define LOGF(x, args...) if (Network) printf(x, ## args)
/* Prints a status message to standard output, unless the server is */
/* running locally as part of an antique or single-player game.     */

struct PLAYER *FirstServer;

int SendSingleHighScore(struct PLAYER *Play,struct HISCORE *Score,
                        int index,char Bold);

void HandleServerMessage(char *buf,struct PLAYER *ReallyFrom) {
/* Given a message "buf" which identifies itself as being from player */
/* "ReallyFrom" by the incoming socket, performs processing and sends */
/* suitable replies.                                                  */
   struct PLAYER *From,*To,*tmp,*pt;
   char Code[BUFLEN],Data[BUFLEN];
   char AICode;
   DopeEntry NewEntry;
   int i;
   long money;
   if (ProcessMessage(buf,&From,&AICode,Code,&To,Data,FirstServer)==-1) {
      LOGF("Bad message\n");
      return;
   }
   if (From!=ReallyFrom && (From!=&Noone || strcmp(Code,"name")!=0)) {
       LOGF("Message is lying about its orgin\n");
       LOGF("%s: %s: %s: %s\n",From ? From->Name : "",Code,
                                 To ? To->Name : "",Data);
       LOGF("Should be from %s\n",ReallyFrom ? ReallyFrom->Name : "NULL");
       return;
   }
   if (strcmp(Code,"msgto")==0) {
      LOGF("%s->%s: %s: %s\n",From->Name,To->Name,Code,Data);
      SendServerMessage(From,AICode,Code,To,Data);
   } else if (strcmp(Code,"name")==0) {
      pt=GetPlayerByName(Data,FirstServer);
      if (pt && pt!=From) {
         SendServerMessage(NULL,C_NONE,"newname",ReallyFrom,NULL);
      } else if ((ReallyFrom && ReallyFrom->Name[0]==0 && Network) || 
                (!Network && From==&Noone)) {
         if (!Network) From=AddPlayer(0,&FirstServer);
         else From=ReallyFrom;
         InitialisePlayer(From);
         strcpy(From->Name,Data);
         LOGF("%s joins the game!\n",From->Name);
         BroadcastToClients(C_NONE,"join",From->Name,NULL,From);
         SendPlayerData(From);
         From->EventNum=E_ARRIVE;
         SendEvent(From);
      } else {
         LOGF("%s will now be known as %s\n",From->Name,Data);
         BroadcastToClients(C_NONE,"rename",Data,From,From);
         strcpy(From->Name,Data);
      }
   } else if (strcmp(Code,"wantquit")==0 && From->EventNum!=E_FINISH) {
      FinishGame(From);
   } else if (strcmp(Code,"requestjet")==0) {
      i=atoi(Data);
      if (From->EventNum==E_ATTACK || From->EventNum==E_DEFEND ||
          From->EventNum==E_WAITFIGHT) {
          BreakoffCombat(From);
      }
      if (From->Turn>=MAXTURNS && From->EventNum!=E_FINISH) {
         FinishGame(From);
      } else if (i!=From->IsAt && From->Turn<MAXTURNS && 
                 From->EventNum==E_NONE) {
         LOGF("%s: Moving to %s\n",From->Name,Location[i].Name);
         From->IsAt=(char)i;
         From->Turn++;
         From->Debt=(long)((float)From->Debt*1.1);
         From->Bank=(long)((float)From->Bank*1.05);
         SendPlayerData(From);
         From->EventNum=1;
         SendEvent(From);
      } else {
         LOGF("%s: DENIED jet to %s\n",From->Name,Location[i].Name);
      }
   } else if (strcmp(Code,"contactspy")==0) {
      for (tmp=FirstServer;tmp;tmp=tmp->Next) {
         if (tmp!=From && GetListEntry(&(tmp->SpyList),From)>=0) {
            SendSpyReport(From,tmp);
         }
      }
   } else if (strcmp(Code,"deposit")==0) {
      money=atol(Data);
      if (From->Bank+money >=0 && From->Cash-money >=0) {
         From->Bank+=money; From->Cash-=money;
         SendPlayerData(From);
      }
   } else if (strcmp(Code,"payloan")==0) {
      money=atol(Data);
      if (From->Debt-money >=0 && From->Cash-money >=0) {
         From->Debt-=money; From->Cash-=money;
         SendPlayerData(From);
      }
   } else if (strcmp(Code,"buyobject")==0) {
      BuyObject(From,Data);
   } else if (strcmp(Code,"answer")==0) {
      HandleAnswer(From,To,Data);
   } else if (strcmp(Code,"done")==0) {
      if (From->EventNum!=E_NONE && From->EventNum<E_OUTOFSYNC) {
         From->EventNum++; SendEvent(From);
      }
   } else if (strcmp(Code,"spyon")==0) {
      if (From->Cash >= SPYPRICE) {
         LOGF("%s now spying on %s\n",From->Name,To->Name);
         From->Cash -= SPYPRICE;
         LoseBitch(From,NULL);
         NewEntry.Play=From; NewEntry.Turns=-1;
         AddListEntry(&(To->SpyList),&NewEntry);
         SendPlayerData(From);
      } else {
         LOGF("%s spy on %s: DENIED\n",From->Name,To->Name);
      }
   } else if (strcmp(Code,"tipoff")==0) {
      if (From->Cash >= TIPPRICE) {
         LOGF("%s tipped off the cops to %s\n",From->Name,To->Name);
         From->Cash -= TIPPRICE;
         LoseBitch(From,NULL);
         NewEntry.Play=From; NewEntry.Turns=0;
         AddListEntry(&(To->TipList),&NewEntry);
         SendPlayerData(From);
      } else {
         LOGF("%s tipoff about %s: DENIED\n",From->Name,To->Name);
      }
   } else if (strcmp(Code,"msg")==0) {
      LOGF("%s: %s\n",From->Name,Data);
      BroadcastToClients(C_NONE,"msg",Data,From,From);
   } else {
      LOGF("%s:%s:%s:%s\n",From->Name,Code,To->Name,Data);
   }
}

void ClientLeftServer(struct PLAYER *Play) {
/* Notifies all clients that player "Play" has left the game and */
/* cleans up after them if necessary.                            */
   BroadcastToClients(C_NONE,"leave",Play->Name,Play,Play);
}

int CheckServerSockets(fd_set *readfs,char WantQuit) {
/* Checks all client players for messages. If the server "wants" to  */
/* quit (i.e. it's interactive, and the user has typed "quit") then  */
/* WantQuit should be set to TRUE. If all the sockets in readfs have */
/* closed and WantQuit is TRUE, then -1 is returned to signal the    */
/* server to quit for good. Otherwise, 0 is returned.                */
   struct PLAYER *tmp;
   char buf[BUFLEN];
   tmp=FirstServer;
   while (tmp) {
      if (FD_ISSET(tmp->fd,readfs)) {
         if (bgets(buf,BUFLEN,tmp->fd)==NULL) {
            if (!WantQuit) {
               LOGF("%s leaves the server!\n",tmp->Name);
               ClientLeftServer(tmp);
            }
            RemovePlayer(tmp,&FirstServer);
            if (!FirstServer && WantQuit) return -1;
            else return 0;
         } else {
            chomp(buf);
            HandleServerMessage(buf,tmp);
         }
      }
      tmp=tmp->Next;
   }
   return 0;
}

void CleanUpServer() {
/* Closes down the server and frees up associated handles and memory */
   while (FirstServer) {
      RemovePlayer(FirstServer,&FirstServer);
   }
   if (Server) close(ListenSock);
}

void BreakHandle(int sig) {
/* Traps an attempt by the user to send dopewars a SIGTERM (e.g. pressing */
/* Ctrl-C) and cleans up the server "properly"                            */
   LOGF("Terminate signal received... cleaning up...\n");
   CleanUpServer();
   exit(1);
}

void PipeHandle(int sig) {
/* Catches and ignores SIGPIPE signals */
   LOGF("SIGPIPE caught\n");
   return;
}

void ServerLoop() {
/* Initialises server, processes network and interactive messages, and */
/* finally cleans up the server on exit.                               */
   struct PLAYER *tmp,*tmp2;
   struct sockaddr_in ServerAddr,ClientAddr;
   fd_set readfs,errorfs;
   int cadsize,topsock,ClientSock;
   char buf[BUFLEN],WantQuit=FALSE;
   char InputClosed=FALSE;
  
   Network=TRUE;
   FirstServer=NULL;
   ListenSock=socket(AF_INET,SOCK_STREAM,0);
   if (ListenSock==-1) {
      perror("create socket"); exit(1);
   }
   fcntl(ListenSock,F_SETFL,O_NONBLOCK);
  
   ServerAddr.sin_family=AF_INET;
   ServerAddr.sin_port=htons(Port);
   ServerAddr.sin_addr.s_addr=INADDR_ANY;
   memset(ServerAddr.sin_zero,0,sizeof(ServerAddr.sin_zero));
   if (bind(ListenSock,(struct sockaddr *)&ServerAddr,
       sizeof(struct sockaddr)) == -1) {
      perror("bind socket"); exit(1);
   }
   printf("Dopewars server version %d.%d.%d ready and waiting for \
connections on port %d\n",VERSION,MAJVER,MINVER,Port);

   if (listen(ListenSock,10)==-1) {
      perror("listen socket"); exit(1);
   }
   if (signal(SIGINT,BreakHandle)==SIG_ERR) {
      printf("Cannot install interrupt handler!\n");
   }
   if (signal(SIGPIPE,PipeHandle)==SIG_ERR) {
      printf("Cannot install pipe handler!\n");
   }
   while (1) {
      FD_ZERO(&readfs);
      if (!InputClosed) FD_SET(0,&readfs);
      FD_SET(ListenSock,&readfs);
      FD_ZERO(&errorfs);
      FD_SET(ListenSock,&errorfs);
      topsock=ListenSock+1;
      for (tmp=FirstServer;tmp;tmp=tmp->Next) {
         FD_SET(tmp->fd,&readfs);
         FD_SET(tmp->fd,&errorfs);
         if (tmp->fd>=topsock) topsock=tmp->fd+1;
      }
      if (select(topsock,&readfs,NULL,&errorfs,NULL)==-1) {
         if (errno==EINTR) continue;
         perror("select"); exit(1);
      }
      if (FD_ISSET(0,&readfs)) {
         if (fgets(buf,sizeof(buf),stdin)==NULL) {
            buf[0]=0;
            puts("Standard input closed.");
            InputClosed=TRUE;
          }
         chomp(buf);
         if (strcmp(buf,"die")==0) {
            puts("Die!");
            break;
         } else if (strcmp(buf,"quit")==0) {
            if (!FirstServer) break;
            WantQuit=TRUE;
            BroadcastToClients(C_NONE,"quit",NULL,NULL,NULL);
         } else if (strncmp(buf,"msg:",4)==0) {
            BroadcastToClients(C_NONE,"msg",buf+4,NULL,NULL);
         } else if (strcmp(buf,"list")==0) {
            if (FirstServer) {
               puts("Users currently logged on:-");
               for (tmp=FirstServer;tmp;tmp=tmp->Next) puts(tmp->Name);
            } else puts("No users currently logged on!");
         } else if (strncmp(buf,"push ",5)==0) {
            tmp=GetPlayerByName(buf+5,FirstServer);
            if (tmp) {
               printf("Pushing %s\n",tmp->Name);
               SendServerMessage(NULL,C_NONE,"push",tmp,NULL);
            } else puts("No such user!");
         } else if (strncmp(buf,"kill ",5)==0) {
            tmp=GetPlayerByName(buf+5,FirstServer);
            if (tmp) {
               printf("%s killed\n",tmp->Name);
               BroadcastToClients(C_NONE,"kill",tmp->Name,tmp,FirstServer);
               RemovePlayer(tmp,&FirstServer);
            } else puts("No such user!");
         }
      }
      if (FD_ISSET(ListenSock,&readfs)) {
         cadsize=sizeof(struct sockaddr);
         if ((ClientSock=accept(ListenSock,(struct sockaddr *)&ClientAddr,
             &cadsize))==-1) {
            perror("accept socket"); exit(1);
         }
         LOGF("server: got connection from %s\n",
                inet_ntoa(ClientAddr.sin_addr));
         tmp=AddPlayer(ClientSock,&FirstServer);
         if (tmp) {
            for (tmp2=FirstServer;tmp2;tmp2=tmp2->Next) {
               if (tmp!=tmp2) {
                  SendServerMessage(NULL,C_NONE,"list",tmp,tmp2->Name);
               }
            }
            SendServerMessage(NULL,C_NONE,"endlist",tmp,NULL);
         }
      } else if (CheckServerSockets(&readfs,WantQuit)==-1) {
         break;
      }
      for (tmp=FirstServer;tmp;tmp=tmp->Next) if (FD_ISSET(tmp->fd,&errorfs)) {
         LOGF("error from client: %d\n",tmp->fd);
         CleanUpServer(); exit(1);
      }
   }
   CleanUpServer();
}

void FinishGame(struct PLAYER *Play) {
/* Tells player "Play" that their time is up... */
   Play->EventNum=E_FINISH;
   SendHighScores(Play);
}

void HighScoreTypeRead(struct HISCORE *HiScore,FILE *fp) {
/* Reads a batch of NUMHISCORE high scores into "HiScore" from "fp" */
   int i;
   char text[100];
   for (i=0;i<NUMHISCORE;i++) {
      if (read_string(fp,HiScore[i].Name,80)==EOF) break;
      read_string(fp,HiScore[i].Time,80);
      read_string(fp,text,80);
      HiScore[i].Money=atol(text);
      HiScore[i].Dead=fgetc(fp);
   }
}

void HighScoreTypeWrite(struct HISCORE *HiScore,FILE *fp) {
/* Writes out a batch of NUMHISCORE high scores from "HiScore" to "fp" */
   int i;
   char text[100];
   for (i=0;i<NUMHISCORE;i++) {
      fwrite(HiScore[i].Name,strlen(HiScore[i].Name)+1,1,fp);
      fwrite(HiScore[i].Time,strlen(HiScore[i].Time)+1,1,fp);
      sprintf(text,"%ld",HiScore[i].Money);
      fwrite(text,strlen(text)+1,1,fp);
      fputc(HiScore[i].Dead,fp);
   }
}

int HighScoreRead(struct HISCORE *MultiScore,struct HISCORE *AntiqueScore) {
/* Reads all the high scores into MultiScore and                           */
/* AntiqueScore (antique mode scores). Returns 1 on success, 0 on failure. */
   FILE *fp;
   memset(MultiScore,0,sizeof(struct HISCORE)*NUMHISCORE);
   memset(AntiqueScore,0,sizeof(struct HISCORE)*NUMHISCORE);
   fp=fopen(HiScoreFile,"r");
   if (fp) {
      HighScoreTypeRead(AntiqueScore,fp);
      HighScoreTypeRead(MultiScore,fp);
      fclose(fp);
   } else return 0;
   return 1;
}

int HighScoreWrite(struct HISCORE *MultiScore,struct HISCORE *AntiqueScore) {
/* Writes out all the high scores from MultiScore and AntiqueScore; returns */
/* 1 on success, 0 on failure.                                              */
   FILE *fp;
   fp=fopen(HiScoreFile,"w");
   if (fp) {
      HighScoreTypeWrite(AntiqueScore,fp);
      HighScoreTypeWrite(MultiScore,fp);
      fclose(fp);
   } else return 0;
   return 1;
}

void SendHighScores(struct PLAYER *Play) {
/* Adds "Play" to the high score list if necessary, and then sends the */
/* scores over the network to "Play"                                   */
   struct HISCORE MultiScore[NUMHISCORE],AntiqueScore[NUMHISCORE],Score;
   struct HISCORE *HiScore;
   struct timeval tv;
   struct tm *timep;
   time_t tim;
   int i,j,InList=-1;
   if (!HighScoreRead(MultiScore,AntiqueScore)) {
      LOGF("Error: Unable to read high score file %s\n",HiScoreFile);
   }
   Score.Money=Play->Cash+Play->Bank-Play->Debt;
   strcpy(Score.Name,Play->Name);
   if (Play->Health==0) Score.Dead=1; else Score.Dead=0;
   if (WantAntique) HiScore=AntiqueScore; else HiScore=MultiScore;
   gettimeofday(&tv,0);
   tim=(time_t)(tv.tv_sec);
   timep=gmtime(&tim);
   strftime(Score.Time,80,"%d-%m-%Y",timep);
   for (i=0;i<NUMHISCORE;i++) {
      if (InList==-1 && (Score.Money > HiScore[i].Money ||
          HiScore[i].Time[0]==0)) {
         InList=i;
         SendPrintMessage(NULL,C_NONE,Play,
                          "Congratulations! You made the high scores!");
         for (j=NUMHISCORE-1;j>i;j--) {
            memcpy(&HiScore[j],&HiScore[j-1],sizeof(struct HISCORE));
         }
         memcpy(&HiScore[i],&Score,sizeof(struct HISCORE));
         break;
      }
   }
   if (InList==-1) {
      SendPrintMessage(NULL,C_NONE,Play,
                       "You didn't even make the high score table...");
   }
   SendServerMessage(NULL,C_NONE,"starthiscore",Play,NULL);

   j=0;
   for (i=0;i<NUMHISCORE;i++) {
      if (SendSingleHighScore(Play,&HiScore[i],j,InList==i)) j++;
   }
   if (InList==-1) SendSingleHighScore(Play,&Score,j,1);
   SendServerMessage(NULL,C_NONE,"endhiscore",Play,NULL);
   if (!HighScoreWrite(MultiScore,AntiqueScore)) {
      LOGF("Error: Unable to write high score file %s\n",HiScoreFile);
   }
}

int SendSingleHighScore(struct PLAYER *Play,struct HISCORE *Score,
                        int index,char Bold) {
/* Sends a single high score in "Score" with position "index" to player  */
/* "Play". If Bold is TRUE, instructs the client to display the score in */
/* bold text.                                                            */
   char Data[100],Price[100];
   if (Score->Time[0]==0) return 0;
   sprintf(Data,"%d^%c %18s  %-14s %-34s %8s ",index,Bold ? 'B' : 'N',
                FormatPrice(Score->Money,Price),Score->Time,Score->Name,
                Score->Dead ? "(R.I.P.)" : "");
   SendServerMessage(NULL,C_NONE,"hiscore",Play,Data);
   return 1;
}

void SendEvent(struct PLAYER *To) {
/* In order for the server to keep track of the state of each client, each    */
/* client's state is identified by its EventNum data member. So, for example, */
/* there is a state for fighting the cops, a state for going to the bank, and */
/* so on. This function instructs client player "To" to carry out the actions */
/* expected of it in its current state. It is the client's responsibility to  */
/* ensure that it carries out the correct actions to advance itself to the    */
/* "next" state; if it fails in this duty it will hang!                       */
   long Money;
   int i,j;
   char text[BUFLEN];
   struct PLAYER *Play;
   if (To->EventNum==E_MAX) To->EventNum=E_NONE;
   if (!To || To->EventNum==E_NONE || To->EventNum>=E_OUTOFSYNC) return;
   Money=To->Cash+To->Bank-To->Debt;

   ClearPrices(To);

   while (To->EventNum<E_MAX) {
      switch (To->EventNum) {
         case E_SUBWAY:
            SendServerMessage(NULL,C_NONE,"subwayflash",To,NULL);
            break;
         case E_COPOFFER:
            for (i=0;i<To->TipList.Number;i++) {
               LOGF("%s: Tipoff from %s\n",To->Name,
                      To->TipList.Data[i].Play->Name);
               To->OnBehalfOf=To->TipList.Data[i].Play;
               SendCopOffer(To,FORCECOPS);
               return;
            }
            for (i=0;i<To->SpyList.Number;i++) {
               LOGF("%s: Checking spy from %s, turns=%d\n",To->Name,
                      To->SpyList.Data[i].Play->Name,
                      To->SpyList.Data[i].Turns);
               if (To->SpyList.Data[i].Turns<0) {
                  LOGF("%s: New spy from %s\n",To->Name,
                         To->SpyList.Data[i].Play->Name);
                  To->OnBehalfOf=To->SpyList.Data[i].Play;
                  SendCopOffer(To,FORCEBITCH);
                  return;
               }
               To->SpyList.Data[i].Turns++;
               if (To->SpyList.Data[i].Turns>3 && 
                   brandom(0,100)<10+To->SpyList.Data[i].Turns) {
                   if (TotalGunsCarried(To) > 0) j=brandom(0,NUMDISCOVER);
                   else j=brandom(0,NUMDISCOVER-1);
                   sprintf(text,"One of your bitches was spying for %s.\
:The spy %s!",To->SpyList.Data[i].Play->Name,Discover[j]);
                   if (j!=DEFECT) LoseBitch(To,NULL);
                   SendPlayerData(To);
                   SendPrintMessage(NULL,C_NONE,To,text);
                   sprintf(text,"Your spy working with %s has been \
discovered!:The spy %s!",To->Name,Discover[j]);
                   if (j==ESCAPE) GainBitch(To->SpyList.Data[i].Play);
                   SendPlayerData(To->SpyList.Data[i].Play);
                   SendPrintMessage(NULL,C_NONE,
                                     To->SpyList.Data[i].Play,text);
                   RemoveListEntry(&(To->SpyList),i);
                   i--;
               }
            }
            if (Money>3000000) i=130;   
            else if (Money>1000000) i=115;   
            else i=100;   
            if (brandom(0,i)>75) {
               if (SendCopOffer(To,NOFORCE)) return;
            }
            break;
         case E_SAYING:
            if (brandom(0,100)<15) SendServerMessage(NULL,C_NONE,
                                                     "saying",To,NULL);
            break;
         case E_LOANSHARK:
            if (To->IsAt==LOANSHARK && To->Debt>0) {
               SendQuestion(NULL,C_ASKLOAN,To,
                            "YN^Would you like to visit the Loan Shark?");
               return;
            }
            break;
         case E_BANK:
            if (To->IsAt==BANK) {
               SendQuestion(NULL,C_ASKBANK,To,
                                 "YN^Would you like to visit the Bank?");
               return;
            }
            break;
         case E_GUNSHOP:
            if (To->IsAt==GUNSHOP && !To->WantAntique) {
               SendQuestion(NULL,C_ASKGUNSHOP,To,
                        "YN^Would you like to visit Dan's House of Guns?");
               return;
            }
            break;
         case E_ROUGHPUB:
            if (To->IsAt==ROUGHPUB) {
               SendQuestion(NULL,C_ASKPUB,To,
                                 "YN^Would you like to visit the pub?");
               return;
            }
            break;
         case E_HIREBITCH:
            if (To->IsAt==ROUGHPUB) {
               To->BitchPrice=brandom(MINBITCHPRICE,MAXBITCHPRICE);
               sprintf(text,"YN^^Would you like to hire a bitch for $%ld?",
                       To->BitchPrice);
               SendQuestion(NULL,C_ASKBITCH,To,text);
               return;
            }
            break;
         case E_ARRIVE:
            for (Play=FirstServer;Play;Play=Play->Next) {
               if (Play!=To && Play->IsAt==To->IsAt) {
                  if (TotalGunsCarried(To)>0) {
                     sprintf(text,"AE^%s is already here!^\
Do you A>ttack, or E>vade?",Play->Name);
                     To->Attacked=Play;
                     SendDrugsHere(To);
                     SendQuestion(NULL,C_NONE,To,text);
                     return;
                  } else {
                     sprintf(text,"%s is already here!",Play->Name);
                     SendPrintMessage(NULL,C_NONE,To,text);
                  }
               }
            }
            SendDrugsHere(To);
            break;
      }
      To->EventNum++;
   } 
   if (To->EventNum >= E_MAX) To->EventNum=E_NONE;
}

int SendCopOffer(struct PLAYER *To,char Force) {
/* In response to client player "To" being in state E_COPOFFER,    */
/* randomly engages the client in combat with the cops or offers   */
/* other random events. Returns 1 if the client should then be     */
/* advanced to the next state, 0 otherwise (i.e. if there are      */
/* questions pending which the client must answer first)           */
/* If Force==FORCECOPS, engage in combat with the cops for certain */
/* If Force==FORCEBITCH, offer the client a bitch for certain      */
   int i;
   long Money;
   i=brandom(0,100);
   if (Force==FORCECOPS) i=100;
   else if (Force==FORCEBITCH) i=0;
   else To->OnBehalfOf=NULL;
   if (i<30) { 
      To->EventNum=E_OFFOBJECT; 
      return(OfferObject(To,Force==FORCEBITCH)); 
   } else if (i<46) { To->EventNum=E_OFFOBJECT; return(RandomOffer(To)); }
   else {
      Money=To->Cash+To->Bank-To->Debt;
      if (Money>3000000) To->Cops=brandom(11,27);
      else if (Money>1000000) To->Cops=brandom(7,14);
      else if (Money>500000) To->Cops=brandom(6,12);
      else if (Money>100000) To->Cops=brandom(2,8);
      else To->Cops=brandom(1,5);
      OfficerHardass(To);
   }
   return 1;
}

void OfficerHardass(struct PLAYER *Play) {
/* Send client "Play" a message announcing the arrival of the cops */
   char text[BUFLEN];
   char *OfficerName[2] = { "Hardass", "Bob" };
   int officer;
   
   if (!Play || Play->EventNum!=E_COPOFFER) return;
   if (Play->Cops==0) { Play->EventNum++; SendEvent(Play); return; }
   officer=(Play->Flags&DEADHARDASS ? 1 : 0);
   if (Play->Cops==1) {
      sprintf(text,"YN^Officer %s is chasing you!",OfficerName[officer]);
   } else {
      sprintf(text,"YN^Officer %s and %d of his deputies are chasing you!",
              OfficerName[officer],Play->Cops-1);
   }
   if (TotalGunsCarried(Play)==0) {
      strcat(text,"^Do you run?");
      SendQuestion(NULL,C_ASKRUN,Play,text);
   } else {
      strcat(text,"^Do you R>un, or F>ight?");
      text[0]='R'; text[1]='F';
      SendQuestion(NULL,C_ASKRUNFIGHT,Play,text);
   }
}

void FinishFightWithHardass(struct PLAYER *Play) {
/* Clean up after a fight between "Play" and the cops. If the cops were */
/* tipped off by another player, inform them of the results.            */
   char text[BUFLEN],buf[BUFLEN];
   if (IsValidPlayer(Play->OnBehalfOf,FirstServer)) {
      LOGF("%s: tipoff by %s finished OK.\n",Play->Name,
             Play->OnBehalfOf->Name);
      RemoveListPlayer(&(Play->TipList),Play->OnBehalfOf);
      sprintf(text,"Following your tipoff, the cops ambushed %s, who ",
              Play->Name);
      if (Play->Health==0) strcat(text,"was shot dead.");
      else {
         sprintf(buf,"escaped with %d bitches.",Play->Bitches);
         strcat(text,buf);
      }
      GainBitch(Play->OnBehalfOf);
      SendPlayerData(Play->OnBehalfOf);
      SendPrintMessage(NULL,C_NONE,Play->OnBehalfOf,text);
   }
   Play->OnBehalfOf=NULL;
   Play->EventNum++; SendEvent(Play);
}

void FireAtHardass(struct PLAYER *Play,char FireType) {
/* Have player "Play" attack the cops.                   */
/* FireType is F_STAND: Player has no gun and didn't run */
/*             F_RUN: Player chose to run                */
/*             F_FIGHT: Player chose to fire back        */
   char text[BUFLEN];
   int Damage,i,j;
   char *OfficerName[2] = { "Hardass", "Bob" };
   int officer;

   if (!Play || Play->EventNum!=E_COPOFFER) return;
   if (Play->Cops==0) { FinishFightWithHardass(Play); return; }
   
   text[0]=0;
   officer=(Play->Flags&DEADHARDASS ? 1 : 0);
   if (FireType==F_STAND) {
      strcpy(text,"^^You stand there like an idiot.");
   } else if (FireType==F_RUN) {
      if (brandom(0,100) > 30+Play->Cops*2) {
         if (Play->Cops==1) strcpy(text,"^^You lose him in the alleys.");
         else strcpy(text,"^^You lose them in the alleys.");
         SendPrintMessage(NULL,C_NONE,Play,text);
         FinishFightWithHardass(Play);
         return;
      } else {
         if (Play->Cops==1) strcpy(text,"^^You can\'t shake him, man!");
         else strcpy(text,"^^You can\'t shake them, man!");
      }
   } else if (FireType==F_FIGHT) {
      Damage=100-brandom(0,Play->Cops)*2;
      for (i=0;i<NUMGUN;i++) for (j=0;j<Play->Carry.Gun[i];j++) {
         Damage+=brandom(0,Gun[i].Damage);
      }
      if (Damage>=100) {
         if (Play->Cops==1) {
            i=brandom(1500,3000);
            sprintf(text,"^^You killed Officer %s! You find $%d on his corpse!",
                    OfficerName[officer],i);
            Play->Cash += i;
            Play->Flags |= DEADHARDASS;
            SendPlayerData(Play);
            SendPrintMessage(NULL,C_NONE,Play,text);
            Play->DocPrice=brandom(1000,2000-5*Play->Health);
            if (brandom(0,100)<75 && Play->DocPrice<=Play->Cash &&
                Play->Health<100) {
               Play->EventNum=E_DOCTOR;
               sprintf(text,"YN^^^^Do you pay a doctor $%d to sew %s up?",
                      Play->DocPrice,
                      (Play->Bitches==0 || Play->WantAntique) ? "you" :
                      "your bitches");
               SendQuestion(NULL,C_ASKSEW,Play,text);
            } else {
               FinishFightWithHardass(Play);
            }
            return;
         } else {
            strcpy(text,"^^You got one, man!");
            Play->Cops--;
         }
      }
   }

   if (Play->Cops==1) strcat(text,"^He\'s"); else strcat(text,"^They\'re");
   strcat(text," firing on you, man! ");
   if (brandom(0,100) > 35-Play->Cops*2) {
      strcat(text,"You've been hit! ");
      Damage=0;
      for (i=0;i<Play->Cops;i++) Damage+=brandom(0,5);
      if (Damage==0) Damage=1;
      if (Damage>Play->Health) {
         if (Play->Bitches==0 || Play->WantAntique) {
            if (Play->Cops==1) strcat(text,"He"); else strcat(text,"They");
            strcat(text," wasted you, man! What a drag!");
            Play->Health=0;
            SendPlayerData(Play);
            SendPrintMessage(NULL,C_NONE,Play,text);
            FinishFightWithHardass(Play);
            return;
         } else {
            strcat(text,"You lost one of your bitches!");
            LoseBitch(Play,NULL);
            Play->Health=100;
         }
      } else {
         Play->Health-=Damage;
      }
   } else {
      if (Play->Cops==1) strcat(text,"He"); else strcat(text,"They");
      strcat(text," missed!");
   }
   SendPlayerData(Play);
   SendPrintMessage(NULL,C_NONE,Play,text);
   OfficerHardass(Play);
}

int RandomOffer(struct PLAYER *To) {
/* Inform player "To" of random offers or happenings. Returns 0 if */
/* the client can immediately be advanced to the next state, or 1  */
/* there are first questions to be answered.                       */
   int r,amount,ind;
   char text[BUFLEN],buf[80];
   clear_bottom();
   r=brandom(0,100);
   if (r<10) {
      strcpy(text,"You were mugged in the subway!");
      To->Cash=To->Cash*(long)brandom(80,95)/100l;
   } else if (r<30) {
      amount=brandom(3,7);
      ind=IsCarryingRandom(To,amount);
      if (ind==-1 && amount>To->CoatSize) return 0;
      strcpy(text,"You meet a friend! ");
      if (ind==-1) {
         ind=brandom(0,NUMDRUG);
         sprintf(buf,"He gives you %d %s.",amount,Drug[ind].Name);
         strcat(text,buf);
         To->Carry.Drug[ind]+=amount;
         To->CoatSize-=amount;
      } else {
         sprintf(text,"You give him %d %s.",amount,Drug[ind].Name);
         To->Carry.Drug[ind]-=amount;
         To->CoatSize+=amount;
      }
      SendPlayerData(To);
      SendPrintMessage(NULL,C_NONE,To,text);
   } else if (r<50) {
      amount=brandom(3,7);
      ind=IsCarryingRandom(To,amount);
      if (ind!=-1) {
         sprintf(text,"Police dogs chase you for %d blocks! You dropped \
some drugs! That's a drag, man!",brandom(3,7));
         To->Carry.Drug[ind]-=amount;
         To->CoatSize+=amount;
         SendPlayerData(To);
         SendPrintMessage(NULL,C_NONE,To,text);
      } else {
         ind=brandom(0,NUMDRUG);
         amount=brandom(3,7);
         if (amount>To->CoatSize) return 0;
         sprintf(text,"You find %d %s on a dead dude in the subway!",
                 amount,Drug[ind].Name);
         To->Carry.Drug[ind]+=amount;
         To->CoatSize-=amount;
         SendPlayerData(To);
         SendPrintMessage(NULL,C_NONE,To,text);
      }
   } else if (r<60 && To->Carry.Drug[WEED]+To->Carry.Drug[HASHISH]>0) {
      ind = (To->Carry.Drug[WEED]>To->Carry.Drug[HASHISH]) ?
            WEED : HASHISH;
      amount=brandom(2,6);
      if (amount>To->Carry.Drug[ind]) amount=To->Carry.Drug[ind];
      sprintf(text,"Your mama made brownies with some of your %s! They \
were great!",Drug[ind].Name);
      To->Carry.Drug[ind]-=amount;
      To->CoatSize+=amount;
      SendPlayerData(To);
      SendPrintMessage(NULL,C_NONE,To,text);
   } else if (r<65) {
      strcpy(text,"YN^There is some weed that smells like paraquat here!");
      strcat(text,"^It looks good! Will you smoke it? ");
      To->EventNum=E_WEED;
      SendQuestion(NULL,C_NONE,To,text);
      return 1;
   } else {
      sprintf(text,"You stopped to %s.",
              StoppedTo[brandom(0,NUMSTOPPEDTO)]);
      amount=brandom(1,10);
      if (To->Cash>=amount) To->Cash-=amount;
      SendPlayerData(To);
      SendPrintMessage(NULL,C_NONE,To,text);
   }
   return 0;
}

int OfferObject(struct PLAYER *To,char ForceBitch) {
/* Offers player "To" bitches/trenchcoats or guns. If ForceBitch is  */
/* TRUE, then a bitch is definitely offered. Returns 0 if the client */
/* can advance immediately to the next state, 1 otherwise.           */
   char text[80];
   int ObjNum;
   if (brandom(0,100)<45 || ForceBitch) {
      if (WantAntique) {
         To->BitchPrice=brandom(MINTRENCHPRICE,MAXTRENCHPRICE);
         sprintf(text,"Would you like to buy a bigger trenchcoat for $%ld?",
                 To->BitchPrice);
      } else {
         To->BitchPrice=brandom(MINBITCHPRICE,MAXBITCHPRICE)/10l;
         sprintf(text,"YN^Hey dude! I'll help carry your drugs for a \
mere $%ld. Yes or no?",To->BitchPrice);
      }
      SendQuestion(NULL,C_ASKBITCH,To,text);
      return 1;
   } else if (TotalGunsCarried(To) < To->Bitches+2) {
      ObjNum=brandom(0,NUMGUN);
      To->GunPrice[ObjNum]=Gun[ObjNum].Price/10;
      if (Gun[ObjNum].Space>To->CoatSize) return 0;
      sprintf(text,"YN^Would you like to buy a %s for $%d ?",
              Gun[ObjNum].Name,To->GunPrice[ObjNum]);
      SendQuestion(NULL,C_ASKGUN,To,text);
      return 1;
   }
   return 0;
}

void SendDrugsHere(struct PLAYER *To) {
/* Sends details of drug prices and drug busts etc. to player "To" */
   int i;
   char msg[BUFLEN],text[BUFLEN],buf[20];
   GenerateDrugsHere(To);
   msg[0]=0;
   for (i=0;i<NUMDRUG;i++) if (To->Here->Deal[i]) {
      if (Drug[i].SpecHigh) {
         if (To->Here->Deal[i]==1) {
            sprintf(text,"Cops made a big %s bust! Prices are outrageous!",
                    Drug[i].Name);
         } else {
            sprintf(text,"Addicts are buying %s at ridiculous prices!",
                    Drug[i].Name);
         }
      } else switch (i) {
         case LUDES:
            strcpy(text,"Rival drug dealers raided a pharmacy and are \
selling cheap ludes!"); break;
         case HASHISH:
            strcpy(text,"The Marrakesh Express has arrived!"); break;
         case WEED:
            strcpy(text,"Columbian freighter dusted the Coast Guard! \
Weed prices have bottomed out!"); break;
         case ACID:
            strcpy(text,"The market is flooded with cheap home-made acid!");
            break;
      } 
      if (msg[0]!=0) strcat(msg,"^");
      strcat(msg,text);
   }
   if (msg[0]) SendPrintMessage(NULL,C_NONE,To,msg);
   text[0]=0; 
   for (i=0;i<NUMDRUG;i++) {
      sprintf(buf,"%d^",To->Here->Price[i]);
      strcat(text,buf);
   }
   SendServerMessage(NULL,C_NONE,"drughere",To,text);
}

void GenerateDrugsHere(struct PLAYER *To) {
/* Generates drug prices and drug busts etc. for player "To" */
   int NumEvents,NumDrugs,i;
   for (i=0;i<NUMDRUG;i++) {
      To->Here->Price[i]=0;
      To->Here->Deal[i]=0;
   }
   NumEvents=0;
   if (brandom(0,100)<70) NumEvents=1;
   if (brandom(0,100)<40 && NumEvents==1) NumEvents=2;
   if (brandom(0,100)<5 && NumEvents==2) NumEvents=3;
   NumDrugs=0;
   while (NumEvents>0) {
      i=brandom(0,NUMDRUG);
      if (To->Here->Deal[i]!=0) continue;
      if (Drug[i].SpecHigh) {
         To->Here->Deal[i]=brandom(1,3);
         To->Here->Price[i]=brandom(Drug[i].MinPrice,Drug[i].MaxPrice)*4;
         NumDrugs++;
         NumEvents--;
      } else if (i==LUDES || i==HASHISH || i==WEED || i==ACID) {
         To->Here->Deal[i]=1;
         To->Here->Price[i]=brandom(Drug[i].MinPrice,Drug[i].MaxPrice)/4;
         NumDrugs++;
         NumEvents--;
      }
   }
   NumDrugs=brandom(Location[(int)To->IsAt].MinDrug,
                    Location[(int)To->IsAt].MaxDrug)-NumDrugs;
   if (NumDrugs >= NUMDRUG) NumDrugs=NUMDRUG;
   while (NumDrugs>0) {
      i=brandom(0,NUMDRUG);
      if (To->Here->Price[i]==0) {
         To->Here->Price[i]=brandom(Drug[i].MinPrice,Drug[i].MaxPrice);
         NumDrugs--;
      }
   }
}

void HandleAnswer(struct PLAYER *From,struct PLAYER *To,char *answer) {
/* Handles the incoming message in "answer" from player "From" and */
/* intended for player "To".                                       */
   char text[BUFLEN],buf[BUFLEN];
   int i;
   if (!From || From->EventNum==E_NONE) return;
/*   LOGF("Answer=%s, EventNum=%d\n",answer,From->EventNum); */
   if (answer[0]=='Y' && From->EventNum==E_OFFOBJECT && From->BitchPrice
       && From->BitchPrice>From->Cash) answer[0]='N';
   if (answer[0]=='Y') switch (From->EventNum) { 
      case E_OFFOBJECT:
         text[0]=0;
         if (From->BitchPrice) {
            sprintf(text,"bitch^0^1");
         } else {
            for (i=0;i<NUMGUN;i++) if (From->GunPrice[i]) {
               sprintf(text,"gun^%d^1",i);
            }
         }
         if (IsValidPlayer(From->OnBehalfOf,FirstServer)) {
            LOGF("%s: offer was on behalf of %s\n",From->Name,
                   From->OnBehalfOf->Name);
            if (From->BitchPrice) {
               sprintf(buf,"%s has accepted your bitch!",From->Name);
               strcat(buf,"^Use the C key to contact your spy.");
               From->OnBehalfOf->Flags |= SPYINGON;
               SendPlayerData(From->OnBehalfOf);
               SendPrintMessage(NULL,C_NONE,From->OnBehalfOf,buf);
               i=GetListEntry(&(From->SpyList),From->OnBehalfOf);
               if (i>=0) From->SpyList.Data[i].Turns=0;
            }
         }
         if (text[0]) BuyObject(From,text);
         From->OnBehalfOf=NULL;
         From->EventNum++; SendEvent(From);
         break;
      case E_LOANSHARK:
         SendServerMessage(NULL,C_NONE,"loanshark",From,NULL);
         break;
      case E_BANK:
         SendServerMessage(NULL,C_NONE,"bank",From,NULL);
         break;
      case E_GUNSHOP:
         for (i=0;i<NUMGUN;i++) From->GunPrice[i]=Gun[i].Price;
         SendServerMessage(NULL,C_NONE,"gunshop",From,NULL);
         break;
      case E_HIREBITCH:
         strcpy(text,"bitch^0^1");
         BuyObject(From,text);
         From->EventNum++; SendEvent(From);
         break;
      case E_ROUGHPUB:
         From->EventNum++; SendEvent(From);
         break;
      case E_WEED:
         SendPrintMessage(NULL,C_NONE,From,
     "You hallucinated for three days on the wildest trip you ever imagined!^\
Then you died because your brain disintegrated!"); 
         From->EventNum++; SendEvent(From);
         break;
      case E_COPOFFER:
         FireAtHardass(From,F_RUN);
         break;
      case E_DOCTOR:
         if (From->Cash >= From->DocPrice) {
            From->Cash -= From->DocPrice;
            From->Health=100;
            SendPlayerData(From);
         }
         From->EventNum++; SendEvent(From);
         break;
   } else if (From->EventNum==E_COPOFFER && 
              (answer[0]=='F' || answer[0]=='R')) {
      FireAtHardass(From,answer[0]=='F' ? F_FIGHT : F_RUN);
   } else if (From->EventNum==E_ARRIVE) {
      if (answer[0]=='A' && IsValidPlayer(From->Attacked,FirstServer)) {
         if (From->Attacked->IsAt==From->IsAt) {
            From->DamageDone=-1;
            if (From->Attacked->EventNum<E_MAX) {
               From->Attacked->ResyncNum=From->Attacked->EventNum;
            }
            From->Attacked->Flags |= FIGHTING;
            SendPlayerData(From->Attacked);
            From->Flags |= FIGHTING;
            SendPlayerData(From);
            From->Attacked->EventNum=E_DEFEND;
            From->ResyncNum=E_ARRIVE+1;
            From->EventNum=E_ATTACK;
            AttackPlayer(From,From->Attacked);
         } else {
            sprintf(text,"Too late - %s has just left!",From->Attacked->Name);
            SendPrintMessage(NULL,C_NONE,From,text);
            From->EventNum++; SendEvent(From);
         }
      } else {
         From->EventNum++; SendEvent(From);
      }
   } else if (From->EventNum==E_ATTACK || From->EventNum==E_DEFEND) {
      if ((answer[0]=='Y' && TotalGunsCarried(From)==0) ||
          (answer[0]=='R' && TotalGunsCarried(From)>0)) {
         BreakoffCombat(From);
      } else {
         From->DamageDone=0;
         AttackPlayer(From,To);
      }
   } else switch(From->EventNum) { 
      case E_ROUGHPUB:
         From->EventNum++;
         From->EventNum++; SendEvent(From);
         break;
      case E_COPOFFER:
         FireAtHardass(From,F_STAND);
         break;
      case E_HIREBITCH: case E_GUNSHOP: case E_BANK: case E_LOANSHARK:
      case E_OFFOBJECT: case E_WEED: case E_DOCTOR:
         if (IsValidPlayer(From->OnBehalfOf,FirstServer)) {
            LOGF("%s: offer was on behalf of %s\n",From->Name,
                   From->OnBehalfOf->Name);
            if (From->BitchPrice && From->EventNum==E_OFFOBJECT) {
               sprintf(text,"%s has rejected your bitch!",From->Name);
               GainBitch(From->OnBehalfOf);
               SendPlayerData(From->OnBehalfOf);
               SendPrintMessage(NULL,C_NONE,From->OnBehalfOf,text);
               RemoveListPlayer(&(From->SpyList),From->OnBehalfOf);
            }   
         }
         From->EventNum++; SendEvent(From);
         break;
   }
}

void BreakoffCombat(struct PLAYER *Attack) {
/* Withdraws from player-player combat that player "Attack" is */
/* currently involved in.                                      */
   struct PLAYER *Defend;
   struct PLAYER *Play,*Victor;
   char FightDone,text[BUFLEN];
   if (!IsValidPlayer(Attack,FirstServer)) {
      LOGF("Players involved in a fight are not valid!\n");
      return;
   }
   if (Attack->EventNum!=E_DEFEND && Attack->EventNum!=E_ATTACK &&
        Attack->EventNum!=E_WAITFIGHT) {
      LOGF("Players in fight are not attack/defending!\n");
      return;
   }
   Victor=NULL;

   if (Attack->EventNum==E_DEFEND) {
      sprintf(text,"%s has got away!",Attack->Name);
      for (Play=FirstServer;Play;Play=Play->Next) {
         if (Play->Attacked==Attack && (Play->EventNum==E_ATTACK ||
             Play->EventNum==E_WAITFIGHT)) {
            Play->Attacked=NULL;
            Play->Flags &= ~FIGHTING;
            SendPlayerData(Play);
            SendPrintMessage(NULL,C_NONE,Play,text);
            Victor=Play;
            Play->EventNum=Play->ResyncNum; SendEvent(Play);
         } 
      }
   } else {
      Victor=Attack;
      Defend=Attack->Attacked;
      if (!IsValidPlayer(Defend,FirstServer)) {
         LOGF("Players involved in a fight are not valid!\n");
         return;
      }
      if (Defend->EventNum!=E_DEFEND) {
         LOGF("Players in fight are not attack/defending!\n");
         return;
      }
 
      Attack->Attacked=NULL;
      FightDone=TRUE;
      for (Play=FirstServer;Play;Play=Play->Next) {
         if (Play->Attacked==Defend && (Play->EventNum==E_ATTACK ||
             Play->EventNum==E_WAITFIGHT)) { FightDone=FALSE; break; }
      }
      sprintf(text,"%s has run off!",Attack->Name);
      SendPrintMessage(NULL,C_NONE,Defend,text);
      if (FightDone) {
         Defend->Flags &= ~FIGHTING;
         SendPlayerData(Defend);
         Defend->EventNum=Defend->ResyncNum; SendEvent(Defend);
      }
   }
   SendPrintMessage(NULL,C_NONE,Attack,
      "Coward! You successfully escaped from the gunfight.");
   Attack->Flags &= ~FIGHTING; SendPlayerData(Attack);
   Attack->EventNum=Attack->ResyncNum; SendEvent(Attack);
}

void AttackPlayer(struct PLAYER *Attack,struct PLAYER *Defend) {
/* Processes a player-player attack from player "Attack" to player "Defend" */
   char text[BUFLEN],buf[BUFLEN];
   int Damage,i,j;

   if (!IsValidPlayer(Attack,FirstServer) || 
       !IsValidPlayer(Defend,FirstServer)) {
      LOGF("Players involved in a fight are not valid!\n");
      return;
   }
   if (Attack->EventNum==E_DEFEND && Defend->EventNum!=E_WAITFIGHT) {
      LOGF("Error: %s has tried to defend against %s, but %s is not \
waiting for return fire!\n",Attack->Name,Defend->Name,Defend->Name);
      return;
   }
   if (Attack->EventNum==E_WAITFIGHT) {
      LOGF("Error: %s has tried to attack %s, but hasn't \
waited for return fire!\n",Attack->Name,Defend->Name);
      return;
   }
   if (Attack->EventNum!=E_DEFEND && Attack->EventNum!=E_ATTACK) {
      LOGF("Error: %s is in wrong state (%d) to attack!",
             Attack->Name,Attack->EventNum);
      return;
   }
   if (TotalGunsCarried(Defend)==0) strcpy(text,"YN");
   else strcpy(text,"FR");

   sprintf(buf,"^%s %sis shooting at you!",Attack->Name,
           Attack->DamageDone==-1 ? "has just arrived and " : "");
   strcat(text,buf);
   Damage=0;
   for (i=0;i<NUMGUN;i++) for (j=0;j<Attack->Carry.Gun[i];j++) {
      Damage+=brandom(0,Gun[i].Damage);
   }
   if (Damage==0) {
      strcat(text,"^You escaped unscathed...");
      sprintf(buf,"You failed to hit %s",Defend->Name);
      SendPrintMessage(Defend,C_NONE,Attack,buf);
   } else {
      strcat(text,"^You've been hit, man!");
      if (Damage>=Defend->Health) {
         if (Defend->Bitches==0) {
            strcat(text," You've been wasted! What a drag!");
            sprintf(buf,"You hit and killed %s!",Defend->Name);
            SendPrintMessage(Defend,C_NONE,Attack,buf);
            SendPlayerData(Defend);
            SendPrintMessage(Attack,C_NONE,Defend,text+3);
            Defend->Health=0;
            BreakoffCombat(Defend); return;
         } else {
            strcat(text," You lost a bitch, man!");
            sprintf(buf,"You killed one of %s's bitches!",Defend->Name);
            SendPrintMessage(Defend,C_NONE,Attack,buf);
            LoseBitch(Defend,NULL);
            Defend->Health=100;
         }
      } else {
         Defend->Health-=Damage;
         sprintf(buf,"You hit %s!",Defend->Name);
         SendPrintMessage(Defend,C_NONE,Attack,buf);
      }
   }
   if (TotalGunsCarried(Defend)==0) {
      strcat(text,"^Do you run for it?");
      text[0]='Y'; text[1]='N';
   } else {
      strcat(text,"^Do you F>ire back, or R>un?");
   }
   LOGF("%s attacks %s\n",Attack->Name,Defend->Name);
   if (Attack->EventNum==E_ATTACK) Attack->EventNum=E_WAITFIGHT;
   else if (Attack->EventNum==E_DEFEND) Defend->EventNum=E_ATTACK;
   SendPlayerData(Defend);
   SendQuestion(Attack,C_NONE,Defend,text);
}

void BuyObject(struct PLAYER *From,char *data) {
/* Processes a request stored in "data" from player "From" to buy an */
/* object (bitch, gun, or drug)                                      */
   char *cp,wrd[BUFLEN],type[BUFLEN];
   int index,i,amount;
   cp=data;
   cp=ExtractWord(type,cp);
   cp=ExtractWord(wrd,cp);
   index=atoi(wrd);
   amount=atoi(cp);
   if (strcmp(type,"drug")==0) {
      if (index>=0 && index<NUMDRUG && From->Carry.Drug[index]+amount >= 0
          && From->CoatSize-amount >= 0 && From->Here->Price[index]!=0
          && From->Cash >= amount*From->Here->Price[index]) {
         From->Carry.Drug[index]+=amount;
         From->CoatSize-=amount;
         From->Cash-=amount*From->Here->Price[index];
         SendPlayerData(From); 
      }
   } else if (strcmp(type,"gun")==0) {
      if (index>=0 && index<NUMGUN && TotalGunsCarried(From)+amount >= 0
          && TotalGunsCarried(From)+amount <= From->Bitches+2
          && From->GunPrice[index]!=0
          && From->CoatSize-amount*Gun[index].Space >= 0
          && From->Cash >= amount*From->GunPrice[index]) {
         From->Carry.Gun[index]+=amount;
         From->CoatSize-=amount*Gun[index].Space;
         From->Cash-=amount*From->GunPrice[index];
         SendPlayerData(From);
      }
   } else if (strcmp(type,"bitch")==0) {
      if (From->Bitches+amount >= 0 
          && From->BitchPrice!=0 
          && amount*From->BitchPrice <= From->Cash) {
         for (i=0;i<amount;i++) GainBitch(From);
         if (amount>0) From->Cash-=amount*From->BitchPrice;
         SendPlayerData(From);
      }
   } 
}

void ClearPrices(struct PLAYER *Play) {
/* Clears the bitch and gun prices stored for player "Play" */
   int i;
   Play->BitchPrice=0;
   for (i=0;i<NUMGUN;i++) Play->GunPrice[0]=0;
}

void GainBitch(struct PLAYER *Play) {
/* Gives player "Play" a new bitch (or larger trenchcoat) */
   Play->CoatSize+=10;
   Play->Bitches++;
}

int LoseBitch(struct PLAYER *Play,Inventory *Inven) {
/* Loses one bitch belonging to player "Play". If drugs or guns are */
/* lost with the bitch, 1 is returned (0 otherwise) and the lost    */
/* items are added to "Inven" if non-NULL                           */
   int losedrug=0,i,num;
   int GunIndex[NUMGUN],tmp;
   if (Inven) ClearInventory(Inven);
   Play->CoatSize-=10;
   if (TotalGunsCarried(Play)>0) {
      if (brandom(0,100)<TotalGunsCarried(Play)*100/(Play->Bitches+2)) {
         for (i=0;i<NUMGUN;i++) GunIndex[i]=i;
         for (i=0;i<NUMGUN*5;i++) {
            num=brandom(0,NUMGUN-1);
            tmp=GunIndex[num+1];
            GunIndex[num+1]=GunIndex[num];
            GunIndex[num]=tmp;
         }
         for (i=0;i<NUMGUN;i++) {
            if (Play->Carry.Gun[GunIndex[i]] > 0) {
               Play->Carry.Gun[GunIndex[i]]--; losedrug=1;
               Play->CoatSize+=Gun[GunIndex[i]].Space;
               if (Inven) Inven->Gun[GunIndex[i]]++;
               break;
            }
         }
      }
   }
   for (i=0;i<NUMDRUG;i++) if (Play->Carry.Drug[i]>0) {
      num=(int)((float)Play->Carry.Drug[i]/(Play->Bitches+2.0)+0.5);
      if (num>0) {
         Play->Carry.Drug[i]-=num;
         if (Inven) Inven->Drug[i]+=num;
         Play->CoatSize+=num;
         losedrug=1;
      }
   }
   while (Play->CoatSize<0) {
      for (i=0;i<NUMDRUG;i++) {
         if (Play->Carry.Drug[i]>0) {
            losedrug=1;
            Play->Carry.Drug[i]--;
            Play->CoatSize++;
            if (Play->CoatSize>=0) break;
         }
      }
   }
   Play->Bitches--;
   return losedrug;
}

